#ifndef GL_H
#define GL_H

#include <stdint.h>

#include "util/region.h"
#include "util/vec3f.h"

struct ASDF_;
struct Edge_;

/*  count_leafs
 *
 *  Returns the number of leaf cells in a tree.
 */
int count_leafs(const struct ASDF_* const asdf);


/*  count_tris
 *
 *  Counts the number of triangles in a tree.
 */
int count_tris(const struct ASDF_* const asdf);


/*  find_bounds
 *
 *  Finds the minimum and maximum vertices in a set of points.
 */
void find_bounds(const float* const p, int count,
                 Vec3f* const min, Vec3f* const max);


/*  find_radius
 *
 *  Finds the farthest distance in a set of vertices from a center point.
 */
float find_radius(const float* const p, int count, Vec3f center);


/*  triangulate
 *
 *  Recursively triangulates an ASDF.  Vertices (and vertex normals)
 *  are stored in the array vdata (with vcount keeping track of how many
 *  floats are stored).
 *
 *  Triangles are stored as indexes into this array, with icount recording
 *  the number of indices saved.
 */
void  triangulate(const struct ASDF_* const asdf,
                  float* const vdata, uint32_t* vcount,
                  uint32_t* const idata, uint32_t* icount,
                  int* const halt);
void _triangulate(const struct ASDF_* const asdf, const Region r,
                  struct Edge_* const cache,
                  float* const vdata, uint32_t* const vcount,
                  uint32_t* const idata, uint32_t* const icount,
                  int* const halt);


/*  evaluate_voxel
 *
 *  Uses marching tetrahedrons to generate a set of triangles
 *  for a given leaf cell.  Triangle vertices in idata as indexes into
 *  the vdata list, and icount is updated appropriately.
 */
void evaluate_voxel(const struct ASDF_* const leaf, const Region r,
                    struct Edge_* const cache,
                    float* const vdata, uint32_t* const vcount,
                    uint32_t* const idata, uint32_t* const icount);


/*  zero_crossing
 *
 *  Finds a zero crossing for a given leaf.  May store new vertices in the
 *  vdata buffer (if we haven't around solved for this crossing).
 *
 *  Returns an index into the vdata buffer.
 */
uint32_t zero_crossing(const struct ASDF_* const leaf, const Region r,
                       struct Edge_* edge_cache[8][8], struct Edge_* const cache,
                       const uint8_t v0, const uint8_t v1,
                       float* const vdata, uint32_t* const vcount);


/*  increase_indices
 *
 *  Increases indices by some value.
 */
void increase_indices(uint32_t* const idata, uint32_t icount, uint32_t di);

#endif