#ifndef NODE_H
#define NODE_H

#include <stdbool.h>

#include "node/results.h"
#include "node/opcodes.h"
#include "util/interval.h"
#include "util/region.h"

/*  Node
 *
 *  Recursive data structure defining a node in a math tree.
 */
typedef struct Node_ {
    Opcode opcode;
    Results  results;
    
    // Rank of the node in the tree.  Constants are -1; X, Y, Z are 0,
    // everything else increases by 1 at each step up the tree.  All nodes
    // are initialized with rank -1.
    int rank;
    
    // Flags
    bool marked;     // general-purpose marking used in tree recursion
    bool constant;   // records whether a node is constant or not

    // Children
    struct Node_* lhs;
    struct Node_* rhs;
    
    // Keep track of where this node was last cloned to.
    struct Node_* clone_address;
} Node;


/*  clone_node
 *  
 *  Clones a single node (non-recursively).
 *  Looks up clone_address of children, which must
 *  be populated with a sane value.
 */
Node* clone_node(Node* n);

////////////////////////////////////////////////////////////////////////////////
// Node constructors
////////////////////////////////////////////////////////////////////////////////

// Binary operations
Node* add_n(Node* left, Node* right);
Node* sub_n(Node* left, Node* right);
Node* mul_n(Node* left, Node* right);
Node* div_n(Node* left, Node* right);

Node* min_n(Node* left, Node* right);
Node* max_n(Node* left, Node* right);
Node* pow_n(Node* left, Node* right);

// Unary arithmetic operators
Node* square_n(Node* n);
Node* sqrt_n(Node* n);
Node* sin_n(Node* n);
Node* cos_n(Node* n);
Node* tan_n(Node* n);
Node* neg_n(Node* n);
Node* buf_n(Node* n);

// Constants
Node* constant_n(float value);

// Variables
Node* X_n(void);
Node* Y_n(void);
Node* Z_n(void);

#endif