#include <stdlib.h>

#include "tree/packed.h"
#include "tree/tree.h"

#include "node/node.h"

PackedTree* make_packed(MathTree* tree)
{
    int num_levels = tree->num_levels;

    PackedTree* packed = malloc(sizeof(PackedTree));
    
    (*packed) = (PackedTree) {
        .nodes      = num_levels ?
                        calloc(num_levels, sizeof(Node**)) : NULL,
        .active     = num_levels ?
                        calloc(num_levels, sizeof(unsigned)) : NULL,
        .disabled   = num_levels ?
                        calloc(num_levels, sizeof(ustack*)) : NULL,
        .num_levels = num_levels,
        .head       = tree->head
    };
 
    // Copy all nodes into the packed tree.
    for (unsigned level=0; level < tree->num_levels; ++level) {
        for (unsigned op=0; op < LAST_OP; ++op) {
            packed->active[level] += tree->active[level][op];
        }
        packed->nodes[level] = malloc(sizeof(Node*)*packed->active[level]);
        int index = 0;
        for (unsigned op=0; op < LAST_OP; ++op) {        
            for (unsigned n=0; n < tree->active[level][op]; ++n) {
                packed->nodes[level][index++] = tree->nodes[level][op][n];
            }
        }
    }
    return packed;
}

void free_packed(PackedTree* packed)
{
    for (unsigned level=0; level < packed->num_levels; ++level) {
        free(packed->nodes[level]);
    }

    
    free(packed->nodes);
    free(packed->active);
    free(packed->disabled);
    
    free(packed);
}

////////////////////////////////////////////////////////////////////////////////

void disable_nodes(PackedTree* tree)
{
    // Mark every node as cacheable.  We'll then go down the tree
    // and mark nodes as uncacheable.
    for (int level=0; level < tree->num_levels; ++level) {
        for (int n=0; n < tree->active[level]; ++n) {
            tree->nodes[level][n]->marked = true;
        }
    }
    tree->head->marked = false;
    
    for (int level=tree->num_levels-1; level >= 0; --level) {
        
        // Save the number of nodes disabled in this pass so that
        // we can reverse the operation later.
        ustack* tmp = tree->disabled[level];
        tree->disabled[level] = malloc(sizeof(ustack));
        *(tree->disabled[level]) = (ustack){0, tmp};
        
        for (int n=0; n < tree->active[level]; ++n) {
            Node* node = tree->nodes[level][n];
            
            // If this node is marked, swap it to the back of the list
            // and decrement the active nodes count.    
            if (node->marked) {
            
                // Fill all of the result slots with this value
                fill_results(node, node->results.i.upper);

                // Figure out where we should swap this node to.
                int back = --tree->active[level];
                Node* swap = tree->nodes[level][back];
                
                // And execute the swap.
                tree->nodes[level][n--]  = swap;
                tree->nodes[level][back] = node;
                
                // Finally, increase the count of disabled nodes
                tree->disabled[level]->count++;
                
                node->marked = false;
            }

            // If this is a max or min node, then we might need to
            // only keep one branch active (if that branch is definitely
            // larger/smaller than the other branch)
            else if (node->opcode == OP_MAX) {
                if (node->lhs->results.i.lower >=
                        node->rhs->results.i.upper) {
                    node->lhs->marked = false;
                } else if (node->rhs->results.i.lower >=
                        node->lhs->results.i.upper) {
                    node->rhs->marked = false;
                } else {
                    node->lhs->marked = false;
                    node->rhs->marked = false;
                }
            } else if (node->opcode == OP_MIN) {
                if (node->lhs->results.i.upper <=
                        node->rhs->results.i.lower) {
                    node->lhs->marked = false;
                } else if (node->rhs->results.i.upper <= 
                        node->lhs->results.i.lower) {
                    node->rhs->marked = false;
                } else {
                    node->lhs->marked = false;
                    node->rhs->marked = false;
                }
            }
            
            // Other node types need to keep both branches active
            else {
                if (node->lhs)  node->lhs->marked = false;
                if (node->rhs)  node->rhs->marked = false;
            }
        }
    }
}
    

void enable_nodes(PackedTree* tree)
{
    for (int level=tree->num_levels-1; level >= 0; --level) {
        tree->active[level] += tree->disabled[level]->count;
        ustack* next = tree->disabled[level]->next;
        free(tree->disabled[level]);
        tree->disabled[level] = next;
    }
}