import ctypes

from   libtree  import libtree
from   interval import Interval

class Mesh(object):
    ''' Mesh objects represent a chunk of indexed geometry.'''
    
    def __init__(self, vdata, idata, X, Y, Z):
        ''' Initializes a mesh object from vertex and index array(s).
        '''
        self.vdata = vdata
        self.idata = idata
        self.vcount = len(vdata)/6
        self.icount = len(idata)
        self.tcount = len(idata)/3
        
        self.X, self.Y, self.Z = X, Y, Z
    
    def save_stl(self, filename):
        libtree.write_stl(self.idata, self.icount, self.vdata, filename)
    
    def save(self, filename):
        libtree.write_mesh(filename, self.X, self.Y, self.Z,
                           self.idata, self.icount,
                           self.vdata, self.vcount)
    
    @classmethod
    def load(self, filename):
        X, Y, Z = Interval(), Interval(), Interval()
        icount, vcount = ctypes.c_uint32(), ctypes.c_uint32()
        libtree.read_mesh_header(filename, X, Y, Z, icount, vcount)
        
        vdata = (ctypes.c_float*(vcount.value*6))()
        idata = (ctypes.c_uint32*icount.value)()
    
        libtree.read_mesh(filename, idata, icount, vdata, vcount)
        
        m = Mesh(vdata, idata, X, Y, Z)
        return m
    
    @classmethod
    def merge(cls, meshes):
        ''' Efficiently combines a set of independent meshes.
            (does not perform vertex deduplication).
        '''
        
        meshes = [m for m in meshes if m is not None]
        if meshes == []:    return None

        
        vdata = [m.vdata for m in meshes]
        idata = [m.idata for m in meshes]
        
        vcount = sum(map(len, vdata))
        icount = sum(map(len, idata))
        
        vdata_merged = (ctypes.c_float*vcount)()
        idata_merged = (ctypes.c_uint32*icount)()
        
        vindex = 0
        iindex = 0
        for v, i in zip(vdata, idata):
            libtree.increase_indices(i, len(i), vindex/6)
            vaddr = ctypes.cast(ctypes.addressof(vdata_merged)+vindex*4,
                                ctypes.c_void_p)
            iaddr = ctypes.cast(ctypes.addressof(idata_merged)+iindex*4,
                                ctypes.c_void_p)
            ctypes.memmove(vaddr, v, len(v)*4)
            ctypes.memmove(iaddr, i, len(i)*4)
            vindex += len(v)
            iindex += len(i)
        
        # Set mesh bounds based on bounds of children
        X = Interval(min(m.X.lower for m in meshes),
                      max(m.X.upper for m in meshes))
        Y = Interval(min(m.Y.lower for m in meshes),
                      max(m.Y.upper for m in meshes))
        Z = Interval(min(m.Z.lower for m in meshes),
                      max(m.Z.upper for m in meshes))
        
        mesh = Mesh(vdata_merged, idata_merged, X, Y, Z)

        return mesh