import  ctypes
import  shutil
import  os
import tempfile
import Queue
from math import ceil, log

from    region      import Region
from    mesh        import Mesh
from    asdf        import ASDF
from    libtree     import libtree
from    interval    import Interval
from    multithread import multithread

################################################################################

class MultiMesh(object):
    ''' Multitresolution mesh object.  Meshes are stored in temporary files
        and can be dynamically loaded when required. '''
    
    def __init__(self, files=[None]*8, id='', tmp=True, bounds=None):
        self.files    = files
        self.id       = id
        self.tmp      = tmp
        print '   '*len(self.id), self.id
        
        if bounds is None:  self.load_bounds()
        else:               self.bounds = bounds
        
        self.branches = [None for i in range(8)]


    def load_bounds(self):
        self.bounds = []
        for f in self.files:
            if f is None:
                self.bounds.append(None)
            else:
                X, Y, Z = Interval(), Interval(), Interval()
                icount, vcount = ctypes.c_uint32(), ctypes.c_uint32()
                libtree.read_mesh_header(f, X, Y, Z, icount, vcount)
                self.bounds.append((X, Y, Z))


    def __del__(self):
        if self.tmp is True:
            for f in self.files:
                if f is not None:
                    os.remove(f)
    
    
    def get_mesh_id(self, frac=0.25):
        ''' Gets a level-of-detail mesh ID based on fill counts.'''
        mesh_id = ''
        for i in range(8):
            if self.files[i] is not None:
                if self.branches[i] and self.fill[i] > frac:
                    mesh_id += self.branches[i].get_mesh_id()
                else:
                    mesh_id += 'M%s_%i,' % (self.id, i)
        return mesh_id
    
    def get_lod_mesh(self, id=None):
        ''' Gets a level-of-detail mesh based on fill counts.'''
        
        if id is None:
            id = ','.join('M%s_%i' % (self.id, i) for i in range(8))
        
        meshes = []
        for i in range(8):
            if 'M%s_%i' % (self.id, i) in id:
                meshes.append(Mesh.load(self.files[i]))
            elif self.branches[i] is not None:
                meshes += self.branches[i].get_lod_mesh(id)
                    
        if self.id == '':
            return Mesh.merge(meshes)
        else:
            return meshes
        
    def assign_indices(self, index=0):
        ''' Gives each leaf cell (i.e. cell without any branches)
            a unique index, returning the next index value to assign.
        '''
        if all(b is None for b in self.branches):
            self.index = index
            return index + 1
        
        self.index = None
        for b in self.branches:
            if b is not None:
                index = b.assign_indices(index)

        return index
            

    def get_vbo_data(self):

        # The number of inside-cube slices
        slices = 4
        
        # Assign each leaf cell a unique index
        if self.id == '':   self.assign_indices()

        # Empty array to fill with vertex positions
        data = []
        
        # Recurse on branches
        if self.index is None:
            for b in self.branches:
                if b is not None:
                    data += b.get_vbo_data()

        # If this is a leaf cell, save its color and vertices
        else:
            
            # Record the color of this cube
            i = self.index + 1
            
            b = (i % 256) / 256.
            i /= 256
            g = (i % 256) / 256.
            i /= 256
            r = (i % 256) / 256.

            rgb = [r, g, b]
            
            # Find maximum bounding box of submeshes
            X = Interval(min(b[0].lower for b in self.bounds if b),
                         max(b[0].upper for b in self.bounds if b))
            Y = Interval(min(b[1].lower for b in self.bounds if b),
                         max(b[1].upper for b in self.bounds if b))
            Z = Interval(min(b[2].lower for b in self.bounds if b),
                         max(b[2].upper for b in self.bounds if b))
            
            # Construct all six faces
            for i in range(slices):
                z = Z.lower + (Z.upper - Z.lower)/(slices-1)*i
                data += [X.lower, Y.lower, z]+rgb # First triangle
                data += [X.lower, Y.upper, z]+rgb
                data += [X.upper, Y.lower, z]+rgb
                data += [X.lower, Y.upper, z]+rgb # Second triangle
                data += [X.upper, Y.upper, z]+rgb
                data += [X.upper, Y.lower, z]+rgb
            for i in range(slices):
                y = Y.lower + (Y.upper - Y.lower)/(slices-1)*i
                data += [X.lower, y, Z.lower]+rgb # First triangle
                data += [X.lower, y, Z.upper]+rgb
                data += [X.upper, y, Z.lower]+rgb
                data += [X.lower, y, Z.upper]+rgb # Second triangle
                data += [X.upper, y, Z.upper]+rgb
                data += [X.upper, y, Z.lower]+rgb
            for i in range(slices):
                x = X.lower + (X.upper - X.lower)/(slices-1)*i
                data += [x, Y.lower, Z.lower]+rgb # First triangle
                data += [x, Y.lower, Z.upper]+rgb
                data += [x, Y.upper, Z.lower]+rgb
                data += [x, Y.lower, Z.upper]+rgb # Second triangle
                data += [x, Y.upper, Z.upper]+rgb
                data += [x, Y.upper, Z.lower]+rgb
        
        # If this was the top-level mesh, smush the data
        # into a ctypes array of floats (so that we can
        # later move it into a VBO)
        if self.id == '':
            self.subcube_count = len(data)/(6*6*3*slices)
            data = (ctypes.c_float*len(data))(*data)

        return data



    def set_fill(self, counts):
        
        if self.index is not None:
            self.fill = [counts[self.index]/8.]*8
        else:
            self.fill = [0]*8
            for i in range(8):
                if self.branches[i] is not None:
                    self.fill[i] = self.branches[i].set_fill(counts)
        return sum(self.fill)
    
    
    def save(self, filename):
        ''' Writes the mesh to a set of files. '''
        if filename[-1] == '/':     filename = filename[:-1]
        
        folder, name = os.path.split(filename)
        name = name.replace('.mesh','')
        
        if not os.path.exists(os.path.join(folder, name+'.mesh')):
            os.makedirs(os.path.join(folder, name+'.mesh'))

        for i in range(8):
            if self.files[i] is None:   continue
            target = os.path.join(folder, name+'.mesh',
                     '%s%s_%i.mesh' % (name, self.id, i))
            shutil.copy(self.files[i], target)
        
        for b in self.branches:
            if b is not None:
                b.save(filename)
    
    @classmethod
    def load(cls, filename, id=''):
        if filename[-1] == '/':     filename = filename[:-1]
        
        folder, name = os.path.split(filename)
        name = name.replace('.mesh','')
        
        # Check to see if we found any meshes at this level of detail
        found = False
        files = [None]*8
        for i in range(8):
            target = os.path.join(folder, name + '.mesh',
                                   '%s%s_%i.mesh' % (name, id, i))
            if os.path.isfile(target):
                files[i] = target
                found = True
            
        # If there were no meshes at this level of detail, we're done
        if not found:   return None
        
        # Create an instance of the class
        m = cls(files, id, tmp=False)
        
        # Otherwise, recurse.
        for i in range(8):
            m.branches[i] = cls.load(filename, id+str(i))
        
        return m

            
################################################################################
            
class MultiASDF(object):
    ''' Multiresolution asdf object.  ASDF structures are stored in
        temporary files and loaded dynamically when required. '''
        
    def __init__(self, file=None, id='', tmp=True):
        self.file     = file
        self.id       = id
        self.branches = [None]*8
        self.tmp      = tmp

    def __del__(self):
        if self.tmp:
            os.remove(self.file)

    def triangulate(self):
        print '  '*len(self.id) + 'Triangulating "%s"' % self.id
        asdf = self.get_asdf()
        
        files = []
        bounds = []
        for b in asdf.branches:
            if b is not None:
                mesh = b.triangulate()
                if mesh is not None:
                    handle, file = tempfile.mkstemp(prefix='mesh_')
                    os.close(handle)
                    files.append(file)
                    bounds.append( (mesh.X.copy(),
                                    mesh.Y.copy(),
                                    mesh.Z.copy()) )
                    mesh.save(files[-1])
                else:
                    files.append(None)
                    bounds.append(None)
            else:
                files.append(None)
        if all(f is None for f in files):
            return None
            
        mm = MultiMesh(files, self.id, bounds=bounds)
        
        mm.branches = []
        for b in self.branches:
            if b is not None:
                mm.branches.append(b.triangulate())
            else:
                mm.branches.append(None)
        return mm

    def count(self):
        return 1 + sum(b.count() for b in self.branches if b is not None)

    def get_asdf(self):
        return ASDF.load(self.file)
    
    @classmethod
    def load(cls, filename, id=''):
        if filename[-1] == '/':     filename = filename[:-1]
        
        folder, name = os.path.split(filename)
        name = name.replace('.asdf','')
        
        target = os.path.join(folder, name + '.asdf', name + id + '.asdf')
        
        if not os.path.isfile(target):
            raise ValueError('Failed to open file %s' % target)
            
        m = cls(target, id, tmp=False)
        
        for i in range(8):
            try:
                m.branches[i] = cls.load(filename, id+str(i))
            except ValueError:
                continue
        
        return m
    
    def save(self, filename):
        folder, name = os.path.split(filename)
        name = name.replace('.asdf','')
        
        if not os.path.exists(os.path.join(folder, name+'.asdf')):
            os.makedirs(os.path.join(folder, name+'.asdf'))
        target = os.path.join(folder, name+'.asdf', name)+self.id+'.asdf'
        
        shutil.copy(self.file, target)
        
        for b in self.branches:
            if b is not None:
                b.save(filename)
    
    
    @classmethod
    def from_vol(cls, filename, offset, mm_per_voxel,
                 ni, nj, nk, region=None, id='', queue=None):

        # If we don't already have a subregion, create a region
        # that fills the space.  The region's voxels are
        # in unshifted units.
        if region is None:
            region = Region(dummy=True)
            region.imin = region.jmin = region.kmin = 0
            region.ni = ni
            region.nj = nj
            region.nk = nk
            region.dx = mm_per_voxel*ni
            region.dy = mm_per_voxel*nj
            region.dz = mm_per_voxel*nk
            libtree.build_arrays(region, 0, 0, 0,
                                 region.dx, region.dy, region.dz)
            region.head = True
        
        voxels = (region.ni-1)*(region.nj-1)*(region.nk-1)
        shift = max(0, int(ceil((log(voxels)/log(2) - 21)/3)))

        handle, file = tempfile.mkstemp(prefix='asdf_')
        os.close(handle)
        m = cls(file, id)

        if queue is None:
            print '  '*len(id),
            print 'Loading region "%s" with shift %i' % (id, shift)


        subregions = region.octsect(all=True, overlap=True)
        # Generate children (running in series)
        if shift > 0:
            for i in range(8):
                if subregions[i] is None:   continue
                m.branches[i] = cls.from_vol(filename, offset, mm_per_voxel,
                                             ni, nj, nk, subregions[i],
                                             id+str(i))

        # Generate a downsampled ASDF for this section
        asdf = ASDF(libtree.import_vol_region(
                          filename, ni, nj, nk,
                          region, shift, offset))
        asdf.scale = 1
        asdf.save(m.file)
        
        if queue is not None:
            queue.put((id, m))
        return m