import ctypes
import os
import math
import threading

from libtree     import libtree
from pixels      import PixelBuffer
from multithread import multithread

################################################################################

def threadsafe(f):
    ''' A decorator that locks the arguments to a function,
        invokes the function, then unlocks the arguments and
        returns.'''
    def wrapped(*args, **kwargs):
        for a in set(list(args) + kwargs.values()):
            if isinstance(a, MathTree):
                a.lock()
        result = f(*args, **kwargs)
        for a in set(list(args) + kwargs.values()):
            if isinstance(a, MathTree):
                a.unlock()
        return result
    return wrapped

def forcetree(f):
    ''' A decorator that forces function arguments to be
        of MathTree type using MathTree.wrap '''
    def wrapped(*args):
        return f(*map(MathTree.wrap, args))
    return wrapped

################################################################################

class MathTree(object):
    ''' A MathTree contains a distance metric math expression.
    
        Arithmetic operators are overloaded to extend the tree with
        either distance metric arithmetic or shape logical expressions,
        depending on the value of the instance variable 'shape'
    '''

    def __init__(self, ptr, shape=False):
        self.ptr    = ptr
        self.shape  = shape
        self.bounds = (float('nan'),) * 6
        
        self._str   = None
        self._lock  = threading.Lock()
        self.locked = False
    
    def __del__(self):
        if libtree is not None:
            libtree.free_tree(self.ptr)
    
    def lock(self):     self._lock.acquire()
    def unlock(self):   self._lock.release()
    
    @staticmethod
    def wrap(value):
        ''' Converts a value to a MathTree if it wasn't already.'''
        if isinstance(value, MathTree):
            return value
        elif type(value) is not float:
            try:
                value = float(value)
            except ValueError:
                raise TypeError('Wrong type for MathTree arithmetic (%s)' %
                                type(value))
        return MathTree.Constant(value)
    

    @staticmethod
    @forcetree
    @threadsafe
    def min(A, B):
        return MathTree(libtree.min_tree(A.ptr, B.ptr))

    @staticmethod
    @forcetree
    @threadsafe
    def max(A, B):
        return MathTree(libtree.max_tree(A.ptr, B.ptr))
        
    @staticmethod
    @forcetree
    @threadsafe
    def pow(A, B):
        return MathTree(libtree.pow_tree(A.ptr, B.ptr))
        
    @staticmethod
    @forcetree
    @threadsafe
    def sqrt(A):
        return MathTree(libtree.sqrt_tree(A.ptr))
        
    #########################
    #  MathTree Arithmetic  #
    #########################
    
    # If shape is set, then + and - perform logical combination;
    # otherwise, they perform arithmetic.
    @forcetree
    @threadsafe
    def __add__(self, rhs):
        if self.shape or rhs.shape:
            return MathTree(libtree.min_tree(self.ptr, rhs.ptr), True)
        else:
            return MathTree(libtree.add_tree(self.ptr, rhs.ptr))

    @forcetree
    @threadsafe
    def __radd__(self, lhs):
        if self.shape or lhs.shape:
            return MathTree(libtree.min_tree(lhs.ptr, self.ptr), True)
        else:
            return MathTree(libtree.add_tree(lhs.ptr, self.ptr))

    @forcetree
    @threadsafe
    def __sub__(self, rhs):
        if self.shape or rhs.shape:
            rhs.unlock() # avoid double-locking rhs
            rhs_ = -rhs
            rhs.lock()
            return MathTree(libtree.max_tree(self.ptr, rhs_.ptr), True)
        else:
            return MathTree(libtree.sub_tree(self.ptr, rhs.ptr))

    @forcetree
    @threadsafe
    def __rsub__(self, lhs):
        if self.shape or lhs.shape:
            self.unlock() # avoid double-locking self
            rhs = -self
            self.lock()
            return MathTree(libtree.max_tree(lhs.ptr, rhs.ptr), True)
        else:
            return MathTree(libtree.sub_tree(lhs.ptr, self.ptr))

    @forcetree
    @threadsafe
    def __and__(self, rhs):
        if self.shape or rhs.shape:
            return MathTree(libtree.max_tree(self.ptr, rhs.ptr), True)
        else:
            raise NotImplementedError(
                '& operator is undefined for non-shape math expressions.')

    @forcetree
    @threadsafe
    def __rand__(self, lhs):
        if self.shape or lhs.shape:
            return MathTree(libtree.max_tree(lhs.ptr, self.ptr), True)
        else:
            raise NotImplementedError(
                '& operator is undefined for non-shape math expressions.')

    @forcetree
    @threadsafe
    def __or__(self, rhs):
        if self.shape or rhs.shape:
            return MathTree(libtree.min_tree(self.ptr, rhs.ptr), True)
        else:
            raise NotImplementedError(
                '| operator is undefined for non-shape math expressions.')

    @forcetree
    @threadsafe
    def __ror__(self, lhs):
        if self.shape or lhs.shape:
            return MathTree(libtree.min_tree(lhs.ptr, self.ptr), True)
        else:
            raise NotImplementedError(
                '| operator is undefined for non-shape math expressions.')

    @forcetree
    @threadsafe
    def __mul__(self, rhs):
        return MathTree(libtree.mul_tree(self.ptr, rhs.ptr))
        
    @forcetree
    @threadsafe
    def __rmul__(self, lhs):
        return MathTree(libtree.mul_tree(lhs.ptr, self.ptr))

    @forcetree
    @threadsafe
    def __div__(self, rhs):
        return MathTree(libtree.div_tree(self.ptr, rhs.ptr))

    @forcetree
    @threadsafe
    def __rdiv__(self, lhs):
        return MathTree(libtree.div_tree(lhs.ptr, self.ptr))

    @forcetree
    @threadsafe
    def __neg__(self):
        return MathTree(libtree.neg_tree(self.ptr))

    ###############################
    ## String and representation ##
    ###############################
    
    def __str__(self):
        if self._str is None:
            self._str = self.make_str()
        return self._str
    
    def make_str(self, verbose=False):
        '''Converts the object into a string by creating a OS pipe,
           instructing the object to print itself into the pipe,
           and reading the output in chunks of maximum size 1024.
        '''
        
        # Create a pipe to get the printout
        read, write = os.pipe()
        
        # Start the print function running in a separate thread
        # (so that we can eat the output and avoid filling the pipe)
        if verbose: printer = libtree.fprint_tree_verbose
        else:       printer = libtree.fprint_tree
        t = threading.Thread(target=printer, args=(self.ptr, write))
        t.daemon = True
        t.start()
        
        s = r = os.read(read, 1024)
        while r:
            r = os.read(read, 1024)
            s += r
        t.join()
        
        os.close(read)
        
        return s
    
    def __repr__(self):
        return "'%s' (tree at %s)" % (self, hex(self.ptr))
    
    def verbose(self):
        return self.make_str(verbose=True)
    
    #################################
    ## Tree manipulation functions ##
    #################################
    @threadsafe
    def map(self, X=None, Y=None, Z=None):

        X = MathTree.wrap(X) if X is not None else None
        Y = MathTree.wrap(Y) if Y is not None else None
        Z = MathTree.wrap(Z) if Z is not None else None

        out = MathTree(libtree.map_tree(self.ptr,
                                        X.ptr if X else None,
                                        Y.ptr if Y else None, 
                                        Z.ptr if Z else None),
                   shape = self.shape)
        return out

    @threadsafe
    def clone(self):
        clone = MathTree(libtree.clone_tree(self.ptr), shape=self.shape)
        return clone
    
    #################################
    ## Rendering functions          #
    #################################

    def render(self, region, threads=8, interrupt=None):

        if interrupt is None:   interrupt = threading.Event()
        halt = ctypes.c_int(0)  # flag to abort render
        
        image = PixelBuffer(region.ni, region.nj)
        
        # Divide the task to share among multiple threads
        clones = [self.clone() for i in range(threads)]
        packed = [libtree.make_packed(c.ptr) for c in clones]
        subregions = region.split_xy(threads)
        args = zip(packed, subregions, [image.pixels]*threads, [halt]*threads)
        
        multithread(libtree.render8, args, interrupt, halt)
        
        for p in packed:    libtree.free_packed(p)
        return image
    
    @staticmethod
    def Constant(f):   return MathTree(libtree.constant_tree(f))
        
    @staticmethod
    def X():    return MathTree(libtree.X_tree())
        
    @staticmethod
    def Y():    return MathTree(libtree.Y_tree())
        
    @staticmethod
    def Z():    return MathTree(libtree.Z_tree())

if libtree:
    X = MathTree.X()
    Y = MathTree.Y()
    Z = MathTree.Z()
